# -*- perl -*-

use Test::Pod tests => 20;

pod_file_ok( './lib/DB/Object.pm' );
pod_file_ok( './lib/DB/Object/Cache/Tables.pm' );
pod_file_ok( './lib/DB/Object/Fields/Field.pm' );
pod_file_ok( './lib/DB/Object/Fields.pm' );
pod_file_ok( './lib/DB/Object/Mysql/Query.pm' );
pod_file_ok( './lib/DB/Object/Mysql/Statement.pm' );
pod_file_ok( './lib/DB/Object/Mysql/Tables.pm' );
pod_file_ok( './lib/DB/Object/Mysql.pm' );
pod_file_ok( './lib/DB/Object/Postgres/Lo.pm' );
pod_file_ok( './lib/DB/Object/Postgres/Query.pm' );
pod_file_ok( './lib/DB/Object/Postgres/Statement.pm' );
pod_file_ok( './lib/DB/Object/Postgres/Tables.pm' );
pod_file_ok( './lib/DB/Object/Postgres.pm' );
pod_file_ok( './lib/DB/Object/Query.pm' );
pod_file_ok( './lib/DB/Object/SQLite/Query.pm' );
pod_file_ok( './lib/DB/Object/SQLite/Statement.pm' );
pod_file_ok( './lib/DB/Object/SQLite/Tables.pm' );
pod_file_ok( './lib/DB/Object/SQLite.pm' );
pod_file_ok( './lib/DB/Object/Statement.pm' );
pod_file_ok( './lib/DB/Object/Tables.pm' );
